''**************************************************************************
'' Implementation of FS API and some maintaining subroutines
''**************************************************************************

#ifndef FS_INX_I_INC
#define FS_INX_I_INC

'' Find File Parameters
string sFfpSearchedFileName$(MAX_LONGNAME_SIZE)
long lFfpLastFoundPos

array asFileIndexTable$( FILE_INDEX_TABLE_SIZE ) of string( FILE_INDEX_SIZE )
#ifdef INDEX_USE_DATA_POOLS
array asFileDataPool$( FILE_INDEX_TABLE_SIZE ) of string( FILE_DATA_BUFFER_SIZE )
#endif

'' Synchronization vars
byte bTimeoutSyncEnabled
long lSyncTimeout
long lSyncTimeoutCounter

''*********************************************************************************
''
'' *** FS API Subroutine ***
'' Calls special subroutines initialising a particular storage medium (f.e.: SmartMedia)
'' that is to be used by the file system.
'' This subroutine retrieves also the parameters of the storage medium.
sub bFileSystemHardwareInit( var byte bpvHdInitOk )
  bTimeoutSyncEnabled = FALSE
  call bHdInit( bpvHdInitOk )
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Initialises internal file system data,
'' reads the boot sector and retrieves current file system settings.
sub bSetupFileSystem( var byte bpvIsFSSetupOk )
  byte blIsFATSetupOk, blIsDirSet

  bTimeoutSyncEnabled = FALSE

  bpvIsFSSetupOk = FALSE
  call vDateVarIni()
#ifdef CORRECT_RTC
  call vInitRtc()
#endif

  call bFATSetup( blIsFATSetupOk )

  if blIsFATSetupOk = TRUE then
    call vInitIndexVars()
    call vInitFileFindVars()
    call vInitSyncVars()
    call bSetCurrentDir( DEF_CURRENTDIR, blIsDirSet )
    if blIsDirSet = TRUE then
      bpvIsFSSetupOk = TRUE
    endif
  endif

  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Creates and returns a new file descriptor for the file named by spFileName$.
'' Initially, the file position indicator for the file is at the beginning of the file.
sub lOpenFile( string spFileName$; long lpFlags; var long lpvHandle )
  byte blFatAttr, blFlagsOk
  long llFileInfoAddr, llFlags

  lpvHandle = -1
  llFlags = lpFlags
  call bCheckFlags( llFlags, blFlagsOk )
  if blFlagsOk = FALSE then
    '' parameter error
    return
  endif
  bTimeoutSyncEnabled = FALSE
  '' look for the filename in the FAT
  call lFATFindFileInfo( spFileName$, 0, llFileInfoAddr )
  '' does the filename exist?
  if llFileInfoAddr < 0 then
    if ( llFlags bitand O_CREAT ) = O_CREAT then
      call bConvertFlagsToFatAttr( llFlags, blFatAttr, DIR_ATTR_FILE )
      call lCreateNewFile( spFileName$, llFlags, DIR_ATTR_FILE, lpvHandle )
    endif
  else
    if llFileInfoAddr >= 0 then
      '' at O_EXIST set handle to zero and return
      if ( llFlags bitand O_EXIST ) = O_EXIST then
        lpvHandle = 0
  	bTimeoutSyncEnabled = TRUE
        return
      endif
      call lOpenExistingFile( llFlags, llFileInfoAddr, lpvHandle )
    endif
  endif
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Closes the file descriptor lpHandle.
sub bCloseFile( long lpHandle; var byte bpvIsFileClosed )
  byte blHandleOk
  long llFiFlags, llFiFatEntryPos, llFiFilePointer, &
    llBufPoolPos, llBufFileStart, llBufFileLen, llLastCluster
  bTimeoutSyncEnabled = FALSE
  bpvIsFileClosed = FALSE
  call bCheckFileHandle( lpHandle, blHandleOk )
  if blHandleOk = TRUE then
    call lParseFileIndex( asFileIndexTable$( lpHandle ), llFiFlags, llFiFatEntryPos,&
      llFiFilePointer, llBufPoolPos, llBufFileStart, llBufFileLen, llLastCluster )
#ifdef INDEX_USE_DATA_POOLS
    asFileDataPool$( llBufPoolPos ) = ""
#endif
    asFileIndexTable$( lpHandle ) = ""
    bpvIsFileClosed = TRUE
'''' #ifdef SM_WRITE_ON_BLOCK_SWITCH
''''     call vInMemBlockFlush()
'''' #endif
    call vSynchronizeFS()
  endif
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Reads up to lpSize bytes from the file with descriptor lpHandle,
'' storing the results in the spvBuffer$. (This is not necessarily
'' a character string, and no terminating null character is added.)
sub lReadFile( long lpHandle; var string spvBuffer$; long lpSize; &
  var long lpvNumBytesRead )
  byte blIsAvailable, blIsOk, blHandleOk, blIsRead
  long llFiFlags, llFiFatEntryPos, llFiFilePointer, llBufPoolPos, &
    llBufFileStart, llBufFileLen, llLastCluster, llPosition, llSize
  bTimeoutSyncEnabled = FALSE
  lpvNumBytesRead = -1
  call bCheckFileHandle( lpHandle, blHandleOk )
  if blHandleOk = TRUE then
    '' if lpSize is negative, return an error; if lpSize is zero, return zero
    if lpSize <= 0 then
      if lpSize = 0 then
        lpvNumBytesRead = 0
      endif
      bTimeoutSyncEnabled = TRUE
      return
    endif
    call lParseFileIndex( asFileIndexTable$( lpHandle ), llFiFlags, llFiFatEntryPos,&
      llFiFilePointer, llBufPoolPos, llBufFileStart, llBufFileLen, llLastCluster )
    llPosition = llFiFilePointer
    '' check write-only flag
    '' combination of O_RDONLY and O_WRONLY is allowed
    if ( llFiFlags bitand O_WRONLY ) = O_WRONLY and &
      ( llFiFlags bitand O_RDONLY ) <> O_RDONLY then
        bTimeoutSyncEnabled = TRUE
        return
    endif
    call bFATGetDirEntry( llFiFatEntryPos, blIsAvailable )
    '' if end-of-file was reached, return zero
    if llPosition = lDirSize then
      lpvNumBytesRead = 0
      bTimeoutSyncEnabled = TRUE
      return
    endif

    llSize = lpSize
    '' correct the size to read
    if lDirSize - llPosition < llSize then
      llSize = lDirSize - llPosition
    endif
    '' check the buffer size
    if llSize > max_len( spvBuffer$ ) then
      llSize = max_len( spvBuffer$ )
    endif

    call bFATReadFile( llFiFatEntryPos, llPosition, spvBuffer$, llSize, blIsRead )

    if blIsRead = TRUE then
      lpvNumBytesRead = llPosition - llFiFilePointer
      call lSetFilePointer( lpHandle, lpvNumBytesRead, FILE_CURRENT, llPosition )
    endif
  endif
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Writes up to lpSize bytes from spBuffer$ to the file with descriptor lpHandle.
'' The data in spBuffer$ is not necessarily a character string and
'' a null character is output like any other character.
sub lWriteFile( long lpHandle; string spBuffer$; long lpSize; var long lpvNumBytesWritten )
  byte blIsRead, blHandleOk, blIsWritten
  long llNewFilePointer, llFiFlags, llFiFatEntryPos, llFiFilePointer, &
    llBufPoolPos, llBufFileStart, llBufFileLen, llLastCluster
  bTimeoutSyncEnabled = FALSE
  lpvNumBytesWritten = -1
  call bCheckFileHandle( lpHandle, blHandleOk )
  if blHandleOk = TRUE then
    '' if lpSize is negative, return an error; if lpSize is zero, return zero
    if lpSize <= 0 then
      if lpSize = 0 then
        lpvNumBytesWritten = 0
      endif
      bTimeoutSyncEnabled = TRUE
      return
    endif
    call lParseFileIndex( asFileIndexTable$( lpHandle ), llFiFlags, llFiFatEntryPos,&
      llFiFilePointer, llBufPoolPos, llBufFileStart, llBufFileLen, llLastCluster )
    '' check read-only flag
    '' combination of O_RDONLY and O_WRONLY is allowed
    if ( llFiFlags bitand O_RDONLY ) = O_RDONLY and &
      ( llFiFlags bitand O_WRONLY ) <> O_WRONLY then
      lpvNumBytesWritten = -1
      bTimeoutSyncEnabled = TRUE
      return
    endif
    llNewFilePointer = llFiFilePointer
    '' at O_APPEND: file pointer must be set to the end of file
    if ( llFiFlags bitand O_APPEND ) = O_APPEND then
      '' read dir entry for the file
      call bFATGetDirEntry( llFiFatEntryPos, blIsRead )
      llNewFilePointer = lDirSize
    endif
    call bFATWriteFile( llFiFatEntryPos, llNewFilePointer, llLastCluster, spBuffer$, lpSize, blIsWritten )
    if blIsWritten = TRUE then
      '' "llFiFilePointer" was eventually modified in "bFATWriteFile", write it back down
      call sBuildFileIndex( asFileIndexTable$( lpHandle ), llFiFlags, llFiFatEntryPos,&
        llNewFilePointer, llBufPoolPos, llBufFileStart, llBufFileLen, llLastCluster )
      lpvNumBytesWritten = llNewFilePointer - llFiFilePointer
    else
      lpvNumBytesWritten = -1
    endif
  endif
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Reads the file position of the file with descriptor lpHandle.
sub lGetFilePointer( long lpHandle; var long lpvCurFilePtr )
  byte blHandleOk
  long llFiFlags, llFiFatEntryPos, llBufPoolPos, llBufFileStart, llBufFileLen, llLastCluster
  bTimeoutSyncEnabled = FALSE
  lpvCurFilePtr = -1
  call bCheckFileHandle( lpHandle, blHandleOk )
  if blHandleOk = TRUE then
    call lParseFileIndex( asFileIndexTable$( lpHandle ), llFiFlags, llFiFatEntryPos,&
      lpvCurFilePtr, llBufPoolPos, llBufFileStart, llBufFileLen, llLastCluster )
  endif
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Changes the file position of the file with descriptor lpHandle.
sub lSetFilePointer( long lpHandle; long lpOffset; byte bpWhence; var long lpvNewFilePtr )
  byte blHandleOk
  long llFiFlags, llFiFatEntryPos, llFiFilePointer, &
    llBufPoolPos, llBufFileStart, llBufFileLen, llLastCluster
  bTimeoutSyncEnabled = FALSE
  lpvNewFilePtr = -1
  call bCheckFileHandle( lpHandle, blHandleOk )
  if blHandleOk = TRUE then
    call lParseFileIndex( asFileIndexTable$( lpHandle ), llFiFlags, llFiFatEntryPos,&
      llFiFilePointer, llBufPoolPos, llBufFileStart, llBufFileLen, llLastCluster )
    call lFATGetNewFilePtr( llFiFatEntryPos, llFiFilePointer, lpOffset, bpWhence, lpvNewFilePtr )
    if lpvNewFilePtr >= 0 then
      llFiFilePointer = lpvNewFilePtr
      llLastCluster = 0
      call sBuildFileIndex( asFileIndexTable$( lpHandle ), llFiFlags, llFiFatEntryPos,&
        llFiFilePointer, llBufPoolPos, llBufFileStart, llBufFileLen, llLastCluster )
    endif
  endif
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Reads the file size of the file with descriptor lpHandle.
sub lGetFileSize( long lpHandle; var long lpvFileSize )
  byte blHandleOk, blIsAvailable
  long llFiFlags, llFiFatEntryPos, llFiFilePointer, &
    llBufPoolPos, llBufFileStart, llBufFileLen, llLastCluster
  bTimeoutSyncEnabled = FALSE
  lpvFileSize = -1
  call bCheckFileHandle( lpHandle, blHandleOk )
  if blHandleOk = TRUE then
    call lParseFileIndex( asFileIndexTable$( lpHandle ), llFiFlags, llFiFatEntryPos,&
      llFiFilePointer, llBufPoolPos, llBufFileStart, llBufFileLen, llLastCluster )
    call bFATGetDirEntry( llFiFatEntryPos, blIsAvailable )
    if blIsAvailable = TRUE then
      lpvFileSize = lDirSize
    endif
  endif
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Creates a new, empty directory with name spFileName$.
sub bCreateDirectory( string spFileName$; var byte bpvIsCreated )
  long llFileInfoAddr, llHandle

  bTimeoutSyncEnabled = FALSE
  bpvIsCreated = TRUE
  '' look for the filename in the FAT
  call lFATFindFileInfo( spFileName$, 0, llFileInfoAddr )
  '' does the filename exist?
  if llFileInfoAddr < 0 then
    call lCreateNewFile( spFileName$, 0, DIR_ATTR_DIRECTORY, llHandle )
    if llHandle = -1 then
      bpvIsCreated = FALSE
    endif
  endif
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Deletes the file or the directory spFileName$.
'' A read-only file (i.e. a file with the set "DIR_ATTR_READONLY" attribute)
'' cannot be removed. A directory must be empty before it can be removed;
'' in other words, it can only contain entries for '.' and '..'.
sub bDeleteFile( string spFileName$; var byte bpvIsDeleted )
  byte blIsOk, blIsEnd, blIsDirEntry
  long llFileInfoAddr, llCurCluster, llNextCluster, llDirAddr

  bTimeoutSyncEnabled = FALSE
  bpvIsDeleted = FALSE
  '' look for the filename in the FAT
  call lFATFindFileInfo( spFileName$, 0, llFileInfoAddr )
  '' does the filename exist?
  if llFileInfoAddr >= 0 then
    call vSynchronizeFS()
    call bFATGetDirEntry( llFileInfoAddr, blIsOk )
    if blIsOk = TRUE then
      '' if the file or the directory is read-only, it can not be deleted
      if  ( bDirAttr bitand DIR_ATTR_READONLY ) = DIR_ATTR_READONLY then
        bTimeoutSyncEnabled = TRUE
        return
      endif
      llNextCluster = wDirCluster
      '' is it a directory?
      if ( bDirAttr bitand DIR_ATTR_DIRECTORY ) = DIR_ATTR_DIRECTORY then
        '' check whether the directory is empty
        call bFATIsDirEmpty( wDirCluster, blIsDirEntry )
        '' if the directory is not empty, leave the sub
        if blIsDirEntry = FALSE then
          bTimeoutSyncEnabled = TRUE
          return
        endif
      endif
      '' erase the chain in the FAT
      blIsEnd = FALSE
      call lFATClusterToSectorAddr( llNextCluster, llDirAddr )
      call vFATMarkSpecDirEntriesAsErased( llDirAddr )
      while blIsEnd = FALSE
        llCurCluster = llNextCluster
	'' the proper clusters are not erased physically
        '' call vFATEraseDataCluster( llCurCluster )
        call lFATGetNextClusterAddr( llCurCluster, 1, llNextCluster )
        call vFATEraseChainCluster( llCurCluster )
	call bFATIsEndCluster( llNextCluster, blIsEnd )
      endwhile
      '' mark dir entry for the file as erased
      call vFATMarkDirEntryAsErased( llFileInfoAddr )
      bpvIsDeleted = TRUE
    endif
  endif
  call vSynchronizeFS()
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Sets Current Directory to the spNewCurrentDir$.
sub bSetCurrentDir( string spNewCurrentDir$; var byte bpvIsDirSet )
  byte blIsPathAbs
  long llFileInfoAddr
  string slFullNewDirName$( MAX_PATH_SIZE )

#ifdef FS_SECURE_INIT
  slFullNewDirName$ = FS_STR_VARS_INIT_VALUE	'$JK
#endif	'' FS_SECURE_INIT

  bTimeoutSyncEnabled = FALSE
  bpvIsDirSet = FALSE
  if spNewCurrentDir$ = ROOT_DIR then
    sCurrentDir$ = spNewCurrentDir$
    lCurrentDirFATAddr = lRootAreaStart
    bpvIsDirSet = TRUE
  else
    slFullNewDirName$ = spNewCurrentDir$
    call bFATIsPathAbsolute( slFullNewDirName$, blIsPathAbs )
    '' if the new current dir path is not absolute then
    '' it is related to the previously set current dir
    if blIsPathAbs = FALSE then
      slFullNewDirName$ = trim$( sCurrentDir$, " ", 0 ) + PATH_LIM_DOS + spNewCurrentDir$
    endif
    '' Look for the filename in the FAT
    call lFATFindFileInfo( slFullNewDirName$, 0, llFileInfoAddr )
    if llFileInfoAddr >= 0 then
      sCurrentDir$ = slFullNewDirName$
      lCurrentDirFATAddr = llFileInfoAddr
      bpvIsDirSet = TRUE
    endif
  endif
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Reads a File Attribute value of the file spFileName$,
'' storing the result in the bpvFileAttr.
sub bGetFileAttributes( string spFileName$; var byte bpvFileAttr; var byte bpvAttrReadOk )
  byte blIsOk
  long llFileInfoAddr

  bTimeoutSyncEnabled = FALSE
  bpvAttrReadOk = FALSE
  '' look for the filename in the FAT
  call lFATFindFileInfo( spFileName$, 0, llFileInfoAddr )
  '' does the filename exist?
  if llFileInfoAddr >= 0 then
    call bFATGetDirEntry( llFileInfoAddr, blIsOk )
    if blIsOk = TRUE then
      bpvFileAttr = bDirAttr
      bpvAttrReadOk = TRUE
    endif
  endif
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Writes a new File Attribute value bpNewFileAttr of the file spFileName$.
sub bSetFileAttributes( string spFileName$; byte bpNewFileAttr; var byte bpvAttrSetOk )
  byte blIsOk
  long llFileInfoAddr

  bTimeoutSyncEnabled = FALSE
  bpvAttrSetOk = FALSE
  '' look for the filename in the FAT
  call lFATFindFileInfo( spFileName$, 0, llFileInfoAddr )
  '' does the filename exist?
  if llFileInfoAddr >= 0 then
    call bFATGetDirEntry( llFileInfoAddr, blIsOk )
    if blIsOk = TRUE then
      bDirAttr = bpNewFileAttr
      call bFATWriteDirEntry( llFileInfoAddr )
      bpvAttrSetOk = TRUE
    endif
  endif
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Retrieves the date and time that a file spFileName$ was created,
'' last accessed, and last modified.
sub bGetFileTime( string spFileName$; var word wpvCreateDate, wpvCreateTime, &
  wpvAccessDate, wpvWriteDate, wpvWriteTime; var byte bpvIsTimeRead )
  byte blIsOk
  long llFileInfoAddr

  bTimeoutSyncEnabled = FALSE
  bpvIsTimeRead = FALSE
  '' look for the filename in the FAT
  call lFATFindFileInfo( spFileName$, 0, llFileInfoAddr )
  '' does the filename exist?
  if llFileInfoAddr >= 0 then
    call bFATGetDirEntry( llFileInfoAddr, blIsOk )
    if blIsOk = TRUE then
      wpvCreateDate = wDirCreateDate
      wpvCreateTime = wDirCreateTime
      wpvAccessDate = wDirAccessDate
      wpvWriteDate = wDirDate
      wpvWriteTime = wDirTime
      bpvIsTimeRead = TRUE
    endif
  endif
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Sets the date and time that a file spFileName$ was created,
'' last accessed, and last modified.
sub bSetFileTime( string spFileName$; word wpCreateDate, wpCreateTime, &
  wpAccessDate, wpWriteDate, wpWriteTime; var byte bpvIsTimeWritten )
  byte blIsOk
  long llFileInfoAddr

  bTimeoutSyncEnabled = FALSE
  bpvIsTimeWritten = FALSE
  '' look for the filename in the FAT
  call lFATFindFileInfo( spFileName$, 0, llFileInfoAddr )
  '' does the filename exist?
  if llFileInfoAddr >= 0 then
    call bFATGetDirEntry( llFileInfoAddr, blIsOk )
    if blIsOk = TRUE then
      wDirCreateDate = wpCreateDate
      wDirCreateTime = wpCreateTime
      wDirAccessDate = wpAccessDate
      wDirDate = wpWriteDate
      wDirTime = wpWriteTime
      call bFATWriteDirEntry( llFileInfoAddr )
      bpvIsTimeWritten = TRUE
    endif
  endif
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Formats a storage media (f.e. SmartMedia) using the settings preferred
'' by the Windows own formatting routines.
sub bFormatMediaLogicalWin( var byte bpvSuccess )
  bTimeoutSyncEnabled = FALSE
  bFATFmtWinFlag = TRUE
  call bFormatMediaLogical( bpvSuccess )
  bFATFmtWinFlag = FALSE
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Formats a storage media (f.e. SmartMedia) using the settings recommended
'' by the SSFDC Forum.
sub bFormatMediaLogical( var byte bpvSuccess )
  long llI, llJ, llNextAddr
  string slSectBuf$( HD_SECT_SIZE )

  bTimeoutSyncEnabled = FALSE
  call bFmtSetBootParameters( bpvSuccess )
  if bpvSuccess = FALSE then
    bTimeoutSyncEnabled = TRUE
    return
  endif

  llNextAddr = 0
  '' Master Boot Sector
  call vFmtSetMBSectorData()
  call sFmtMakeMBSector( slSectBuf$ )
  call vHdWriteString( llNextAddr, 0, HD_SECT_SIZE, slSectBuf$ )
  llNextAddr = llNextAddr + HD_SECT_SIZE

  '' Reserved Area
  slSectBuf$ = fill$( "00"%, HD_SECT_SIZE )
  for llI = 1 to wBsaHiddenSectors - 1
    call vHdWriteString( llNextAddr, 0, HD_SECT_SIZE, slSectBuf$ )
    llNextAddr = llNextAddr + HD_SECT_SIZE
  next

  '' PartitionBootSector
  call vFmtSetPBSectData()
  call sFmtMakePBSector( slSectBuf$ )
  call vHdWriteString( llNextAddr, 0, HD_SECT_SIZE, slSectBuf$ )
  llNextAddr = llNextAddr + HD_SECT_SIZE

  '' Reserved Sector
  '' == unused ==

  '' FAT Area
  for llI = 0 to bBpbNumberOfFATs - 1
    slSectBuf$ = fill$( "00"%, HD_SECT_SIZE )
    slSectBuf$ = stos$( slSectBuf$, 0, "f8 ff ff"%, 3 )
    if remove$( sBsFileSystem$, " " ) = FAT16_NAME then
      slSectBuf$ = ntos$( slSectBuf$, 3, 1, 0ffh )
    endif

    call vHdWriteString( llNextAddr, 0, HD_SECT_SIZE, slSectBuf$ )
    llNextAddr = llNextAddr + HD_SECT_SIZE

    slSectBuf$ = stos$( slSectBuf$, 0, "00 00 00 00"%, 4 )
    for llJ = 1 to wBpbSectorsPerFAT - 1
      call vHdWriteString( llNextAddr, 0, HD_SECT_SIZE, slSectBuf$ )
      llNextAddr = llNextAddr + HD_SECT_SIZE
    next
  next

  '' Directory Entry
  for llI = 0 to ( ( wBpbRootEntries * 32 ) / HD_SECT_SIZE )- 1
    call vHdWriteString( llNextAddr, 0, HD_SECT_SIZE, slSectBuf$ )
    llNextAddr = llNextAddr + HD_SECT_SIZE
  next

#ifdef SM_WRITE_ON_BLOCK_SWITCH
  call vInMemBlockFlush()
#endif
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Searches the "spSearchedFileName$" file name (wildcards accepted)
sub bFindFirstFile( string spSearchedFileName$; var string spvFfdStruct$; var byte bpvFound )
  bTimeoutSyncEnabled = FALSE
  bpvFound = FALSE
  if max_len( spvFfdStruct$ ) < FFD_STRUCT_SHORT_SIZE then
    bTimeoutSyncEnabled = TRUE
    return
  endif
  lFfpLastFoundPos = 0
  sFfpSearchedFileName$ = spSearchedFileName$
  call bSearchForFfpFile( spvFfdStruct$, bpvFound )
  if bpvFound = FALSE then
    sFfpSearchedFileName$ = ""
  endif
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Searches the same file name (wildcards accepted) as in the previous call of
'' the "bFindFirstFile" subroutine once more
sub bFindNextFile( var string spvFfdStruct$; var byte bpvFound )
  bTimeoutSyncEnabled = FALSE
  bpvFound = FALSE
  if max_len( spvFfdStruct$ ) < FFD_STRUCT_SHORT_SIZE then
    bTimeoutSyncEnabled = TRUE
    return
  endif
  if sFfpSearchedFileName$ = "" then
    bTimeoutSyncEnabled = TRUE
    return
  endif
  lFfpLastFoundPos = lFfpLastFoundPos + DIR_ENTRY_SIZE
  call bSearchForFfpFile( spvFfdStruct$, bpvFound )
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Reads the parameters of the currently connected storage media
sub bGetHardwareInfo( var string spvInfoSet$; var byte bpvIsRead )
  bTimeoutSyncEnabled = FALSE
  call bSMGetInfoSet( spvInfoSet$, bpvIsRead )
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Reads the FAT boot record from a formatted storage media
sub bGetFileSystemInfo( var string spvBootRecord$; var byte bpvIsBootRecRead )
  bTimeoutSyncEnabled = FALSE
  call bFATReadBootRecordToString( spvBootRecord$, bpvIsBootRecRead )
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Calculate the size of free space on storage media
sub lGetDiskFreeSpace( var long lpvFreeSpace )
  bTimeoutSyncEnabled = FALSE
  call lFATGetDiskFreeSpace( lpvFreeSpace )
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Writes to the media all FAT structures temporary saved in the RAM
sub vSynchronizeFS()
  bTimeoutSyncEnabled = FALSE
#ifdef FAT_IN_RAM
  call vFATSynchronizeTable()
#ifdef DIR_ENTRIES_IN_RAM
  call vFATSynchronizeDirEntries()
#endif
#ifdef SM_WRITE_ON_BLOCK_SWITCH
  call vInMemBlockFlush()
#endif
#endif
  bTimeoutSyncEnabled = TRUE
end


'' *** FS API Subroutine ***
'' Reads the recently set synchronization timeout value and
'' the current value of the timeout counter
sub lGetSyncTimeout( var long lpvSyncTimeout; var long lpvCurSyncTimeoutCounter )
  lpvSyncTimeout = lSyncTimeout
  lpvCurSyncTimeoutCounter = diff_ticks( lSyncTimeoutCounter ) / MSEC_PER_SECOND
end


'' *** FS API Subroutine ***
'' Sets the new synchronization timeout value and
'' reads the previously set synchronization timeout value
sub lSetSyncTimeout( long lNewSyncTimeout; var long lpvPrevSyncTimeout )
  lpvPrevSyncTimeout = lSyncTimeout
  lSyncTimeout = lNewSyncTimeout
  lSyncTimeoutCounter = ticks()
end


'' *** FS API Subroutine ***
'' Calls the "vSynchronizeFS" subroutine when the synchronization timeout is over
sub bSynchronizeFSRegularly( var byte bpvTimeoutReached )
  bpvTimeoutReached = FALSE
  if bTimeoutSyncEnabled = TRUE and lSyncTimeout > 0 then
    bTimeoutSyncEnabled = FALSE
    if diff_ticks( lSyncTimeoutCounter ) / MSEC_PER_SECOND >= lSyncTimeout then
      disable_tsw
      call vSynchronizeFS()
      enable_tsw
      lSyncTimeoutCounter = ticks()
      bpvTimeoutReached = TRUE
    endif
    bTimeoutSyncEnabled = TRUE
  endif
end
''
''*********************************************************************************


''*********************************************************************************
''
sub vInitIndexVars()
  long llI
  for llI = 0 to FILE_INDEX_TABLE_SIZE - 1
    asFileIndexTable$( llI ) = ""
#ifdef INDEX_USE_DATA_POOLS
    asFileDataPool$( llI ) = ""
#endif
  next
end


sub sBuildFileIndex( var string spvFileIndex$; long lpFiFlags, &
  lpFiFatEntryPos, lpFiFilePointer, lpBufPoolPos, lpBufFileStart, lpBufFileLen, lpLastCluster )
  spvFileIndex$ = ntos$( spvFileIndex$, FI_FLAGS_INX, FI_FLAGS_SIZE, lpFiFlags )
  spvFileIndex$ = ntos$( spvFileIndex$, FI_FAT_ENTRY_POS_INX, FI_FAT_ENTRY_POS_SIZE, lpFiFatEntryPos )
  spvFileIndex$ = ntos$( spvFileIndex$, FI_FILE_POINTER_INX, FI_FILE_POINTER_SIZE, lpFiFilePointer )
  spvFileIndex$ = ntos$( spvFileIndex$, FI_BUF_POOL_POS_INX, FI_BUF_POOL_POS_SIZE, lpBufPoolPos )
  spvFileIndex$ = ntos$( spvFileIndex$, FI_BUF_FILE_START_INX, FI_BUF_FILE_START_SIZE, lpBufFileStart )
  spvFileIndex$ = ntos$( spvFileIndex$, FI_BUF_FILE_LEN_INX, FI_FILE_POINTER_SIZE, lpBufFileLen )
  spvFileIndex$ = ntos$( spvFileIndex$, FI_LAST_CLUSTER_INX, FI_LAST_CLUSTER_SIZE, lpLastCluster )
end


sub lParseFileIndex( var string spvFileIndex$; var long lpvFiFlags, &
  lpvFiFatEntryPos, lpvFiFilePointer, lpvBufPoolPos, lpvBufFileStart, lpvBufFileLen, lpvLastCluster )
  lpvFiFlags = nfroms( spvFileIndex$, FI_FLAGS_INX, FI_FLAGS_SIZE )
  lpvFiFatEntryPos = nfroms( spvFileIndex$, FI_FAT_ENTRY_POS_INX, FI_FAT_ENTRY_POS_SIZE )
  lpvFiFilePointer = nfroms( spvFileIndex$, FI_FILE_POINTER_INX, FI_FILE_POINTER_SIZE )
  lpvBufPoolPos = nfroms( spvFileIndex$, FI_BUF_POOL_POS_INX, FI_BUF_POOL_POS_SIZE )
  lpvBufFileStart = nfroms( spvFileIndex$, FI_BUF_FILE_START_INX, FI_BUF_FILE_START_SIZE )
  lpvBufFileLen = nfroms( spvFileIndex$, FI_BUF_FILE_LEN_INX, FI_FILE_POINTER_SIZE )
  lpvLastCluster = nfroms( spvFileIndex$, FI_LAST_CLUSTER_INX, FI_LAST_CLUSTER_SIZE )
end


sub lGetFileIndexBufParams( string spvFileIndex$; var long lpvBufPoolPos, &
  lpvBufFileStart, lpvBufFileLen )
  lpvBufPoolPos = nfroms( spvFileIndex$, FI_BUF_POOL_POS_INX, FI_BUF_POOL_POS_SIZE )
  lpvBufFileStart = nfroms( spvFileIndex$, FI_BUF_FILE_START_INX, FI_BUF_FILE_START_SIZE )
  lpvBufFileLen = nfroms( spvFileIndex$, FI_BUF_FILE_LEN_INX, FI_FILE_POINTER_SIZE )
end
''
''*********************************************************************************


sub vInitSyncVars()
  lSyncTimeout = -1
  lSyncTimeoutCounter = -1
  bTimeoutSyncEnabled = TRUE
end


sub bCheckFileHandle( long lpHandle; var byte bpvHandleOk )
  bpvHandleOk = FALSE
  if lpHandle >= 0 and lpHandle < FILE_INDEX_TABLE_SIZE then
    if asFileIndexTable$( lpHandle ) <> "" then
      bpvHandleOk = TRUE
    endif
  endif
end


'' converts file flags used in the "open" function call
'' to the FAT file attributes
sub bConvertFlagsToFatAttr( long lpOpenFlags; var byte bpvFatAttr; byte bpType )
  if bpType = DIR_ATTR_FILE then
    bpvFatAttr = DIR_ATTR_FILE bitor DIR_ATTR_ARCHIVE
    '' combination of O_RDONLY and O_WRONLY is allowed
    if ( lpOpenFlags bitand O_RDONLY ) = O_RDONLY and &
      ( lpOpenFlags bitand O_WRONLY ) <> O_WRONLY then
        bpvFatAttr = bpvFatAttr bitor DIR_ATTR_READONLY
    endif
  else
    if bpType = DIR_ATTR_DIRECTORY then
      bpvFatAttr = DIR_ATTR_DIRECTORY
    endif
  endif
end


sub bCheckFlags( var long lpvFlags; var byte bpvFlagsOk )
  if ( lpvFlags bitand O_FAM_MASK ) = O_FAM_MASK then
    '' a file access mode of zero is the same as O_RDWR
    lpvFlags = lpvFlags bitor O_RDWR
  else
    if ( lpvFlags bitand O_RDWR ) = O_RDWR then
      if ( lpvFlags bitand O_RDONLY ) = O_RDONLY then
        lpvFlags = lpvFlags bitand bitnot(O_RDONLY)
      endif
      if ( lpvFlags bitand O_WRONLY ) = O_WRONLY then
        lpvFlags = lpvFlags bitand bitnot(O_RDONLY)
      endif
    endif
  endif
  bpvFlagsOk = TRUE
end


sub lCreateNewFile( string spFileName$; long lpFlags; byte bpAttr; var long lpvHandle )
  byte blIsAvailable, blIsInRootDir, blIsCurDirRoot, blIsPathAbs
  long llNextPathPos, llDirAddr, llAddrInCluster, llFileInfoAddr, llDirAttr, &
    llPrevPathPos, llSectAddr, llNewFileInfoAddr, llPrevDirAddr, llTempPos
  string slNextPath$(MAX_LONGNAME_SIZE)
  string slLongFileName$(MAX_LONGNAME_SIZE)

#ifdef FS_SECURE_INIT
  slNextPath$ = FS_STR_VARS_INIT_VALUE		'$JK
  slLongFileName$ = FS_STR_VARS_INIT_VALUE	'$JK
#endif	'' FS_SECURE_INIT

  llNextPathPos = 0
  blIsAvailable = TRUE
  llDirAddr = lCurrentDirFATAddr
  llPrevPathPos = 0
  llPrevDirAddr = -1
  call sFATGetNextPathName( spFileName$, llNextPathPos, slNextPath$, PATH_NAME_TYPE_SHORT )
  call bFATIsPathAbsolute( spFileName$, blIsPathAbs )
  if blIsPathAbs = TRUE or lCurrentDirFATAddr = lRootAreaStart then
    blIsInRootDir = TRUE
  else
    llFileInfoAddr = lCurrentDirFATAddr
    blIsInRootDir = FALSE
  endif
  while llNextPathPos < len(spFileName$) and blIsAvailable = TRUE
    call lFATFindFileInfo( slNextPath$, llDirAddr, llFileInfoAddr )
    if llFileInfoAddr = -1 then
      llNextPathPos = llPrevPathPos
      call sFATGetNextPathName( spFileName$, llNextPathPos, slLongFileName$, PATH_NAME_TYPE_LONG )
      blIsCurDirRoot = FALSE
      if sCurrentDir$ = ROOT_DIR then
        blIsCurDirRoot = TRUE
      endif
      if llPrevPathPos = 0 and blIsCurDirRoot = TRUE then
        llFileInfoAddr = llDirAddr
        call bFATMakeNewDirEntry( llFileInfoAddr, slNextPath$, slLongFileName$, DIR_ATTR_DIRECTORY, blIsAvailable )
      else
        llFileInfoAddr = llPrevDirAddr
        call bFATGetDirEntry( llFileInfoAddr, blIsAvailable )
        if blIsAvailable = TRUE then
          call bFATMakeNewDirEntry( llFileInfoAddr, slNextPath$, slLongFileName$, DIR_ATTR_DIRECTORY, blIsAvailable )
        endif
      endif
      if blIsAvailable = TRUE then
        call lFATFindFileInfo( slNextPath$, llDirAddr, llFileInfoAddr )
        if llFileInfoAddr < 0 then
          blIsAvailable = FALSE
        endif
      endif
    endif
    llPrevPathPos = llNextPathPos
    if blIsAvailable = TRUE then
      call bFATGetDirEntry( llFileInfoAddr, blIsAvailable )
      if blIsAvailable = TRUE then
        llDirAddr = wDirCluster
        call lFATClusterToSectorAddr( llDirAddr, llSectAddr )
        llDirAddr = llSectAddr
        llPrevDirAddr = llFileInfoAddr
      endif
      call sFATGetNextPathName( spFileName$, llNextPathPos, slNextPath$, PATH_NAME_TYPE_SHORT )
    endif
  endwhile

  if blIsAvailable = TRUE then
    llTempPos = llPrevPathPos
    call sFATGetNextPathName( spFileName$, llTempPos, slLongFileName$, PATH_NAME_TYPE_LONG )
    if blIsInRootDir = TRUE then
      llFileInfoAddr = llDirAddr
      call bFATMakeNewDirEntry( llFileInfoAddr, slNextPath$, slLongFileName$, bpAttr, blIsAvailable )
    else
      call bFATGetDirEntry( llFileInfoAddr, blIsAvailable )
      if blIsAvailable = TRUE then
        call bFATMakeNewDirEntry( llFileInfoAddr, slNextPath$, slLongFileName$, bpAttr, blIsAvailable )
      endif
    endif
    '' was new dir entry made?
    if blIsAvailable = TRUE then
      '' is a file (not a directory) concerned?
      if ( bpAttr bitand DIR_ATTR_DIRECTORY ) <> DIR_ATTR_DIRECTORY then
        if blIsPathAbs = FALSE then
          '' build absolute path
          slNextPath$ = sCurrentDir$ + PATH_LIM_DOS + spFileName$
        else
          slNextPath$ = spFileName$
        endif
        llFileInfoAddr = 0
        call lFATFindFileInfo( slNextPath$, llFileInfoAddr, llNewFileInfoAddr )
        llFileInfoAddr = llNewFileInfoAddr
        call lOpenExistingFile( lpFlags, llFileInfoAddr, lpvHandle )
      else
        lpvHandle = 0
      endif
    endif
  endif
  call vSynchronizeFS()
end


sub lOpenExistingFile( long lpFlags; long lpFileInfoAddr; var long lpvHandle )
  byte blIsRead
  long llIndex, llBufFileLen, llFilePointer, llLastCluster, llFirstCluster
  string slFileIndex$( FILE_INDEX_SIZE )

#ifdef FS_SECURE_INIT
  slFileIndex$ = FS_STR_VARS_INIT_VALUE	'$JK
#endif	'' FS_SECURE_INIT

  llIndex = 0
  '' look for the first empty entry in the file index table (open file table)
  while llIndex < FILE_INDEX_TABLE_SIZE
    if asFileIndexTable$( llIndex ) = "" then
      lpvHandle = llIndex
      llIndex = FILE_INDEX_TABLE_SIZE
    else
      llIndex = llIndex + 1
    endif
  endwhile
  '' read dir entry for the file
  call bFATGetDirEntry( lpFileInfoAddr, blIsRead )
  '' is there any free place in the file index table?
  if lpvHandle <> -1 then
    '' the size of temporary buffer for file data depends also on the file size
    llBufFileLen = FILE_DATA_BUFFER_SIZE
    if llBufFileLen > lDirSize then
      llBufFileLen = lDirSize
    endif
    llFilePointer = 0
    '' at O_APPEND set file pointer to last position of file
    if ( lpFlags bitand O_APPEND ) = O_APPEND then
      llFilePointer = lDirSize
      llFirstCluster = wDirCluster
      call lFATGetLastClusterInChain( llFirstCluster, llLastCluster )
    else
      llLastCluster = wDirCluster
    endif
    '' build a new index in the file index table;
    '' the place for data buffer is reserved on the file data pool,
    '' the buffer may be accessed by file handle
    call sBuildFileIndex( slFileIndex$, lpFlags, lpFileInfoAddr, llFilePointer, &
      lpvHandle, 0, llBufFileLen, llLastCluster )
    '' insert the new index into the file index table
    asFileIndexTable$( lpvHandle ) = slFileIndex$
#ifdef INDEX_USE_DATA_POOLS
    if lDirSize > 0 then
      call bReadFileToDataPool( lpvHandle, 0, blIsRead )
    endif
#endif
  endif
end


#cm
#ifdef INDEX_USE_DATA_POOLS
sub bReadFileToDataPool( long lpHandle, lpAddrInFile; var byte bpvIsRead )
  byte blIsAvailable, blNeedNextCluster, blIsOk
  long llFiFlags, llFiFatEntryPos, llFiFilePointer, llBufPoolPos, llBufFileStart, llBufFileLen
  long llCurCluster, llNextCluster, llCurAddr, llAddrCounter, llSizeToRead, llReadCounter
  string slBuffer$(FILE_DATA_BUFFER_SIZE)

  bpvIsRead = FALSE

  call lParseFileIndex( asFileIndexTable$( lpHandle ), llFiFlags, llFiFatEntryPos,&
    llFiFilePointer, llBufPoolPos, llBufFileStart, llBufFileLen )

  call bFATGetDirEntry( llFiFatEntryPos, blIsAvailable )
  if blIsAvailable = FALSE then
    return
  endif

  '' is the read address not over the file size ?
  if lpAddrInFile < lDirSize then
    blIsAvailable = FALSE
    llReadCounter = 0
    llCurCluster = wDirCluster
    while llBufFileLen > llReadCounter
      blNeedNextCluster = FALSE
      '' look up the cluster chain to find to which cluster "lpAddrInFile" belongs
      call lFATFindSectorByAddr( lpAddrInFile, llCurCluster, llCurAddr )

      if llCurAddr <> -1 then
        '' are all needed bytes in the current cluster ?
        if lBytesPerCluster - mod( lpAddrInFile, lBytesPerCluster ) >= llBufFileLen - llReadCounter then
          llSizeToRead = llBufFileLen - llReadCounter
        else
          llSizeToRead = lBytesPerCluster - mod( lpAddrInFile, lBytesPerCluster )
          blNeedNextCluster = TRUE
        endif

        llCurAddr = llCurAddr + mod( lpAddrInFile + llReadCounter, lBytesPerCluster )
        call sHdReadString( llCurAddr, llSizeToRead, slBuffer$, blIsOk )
        asFileDataPool$( llBufPoolPos ) = stos$( asFileDataPool$( llBufPoolPos ), &
          llReadCounter, slBuffer$, llSizeToRead )
        bpvIsRead = TRUE

        llReadCounter = llReadCounter + llSizeToRead
        if blNeedNextCluster = TRUE then
          call lFATGetNextClusterAddr( llCurCluster, llNextCluster )
          llCurCluster = llNextCluster
        endif
      endif
    endwhile
    if bpvIsRead = TRUE then
      call sBuildFileIndex( asFileIndexTable$( lpHandle ), llFiFlags, llFiFatEntryPos,&
    	llFiFilePointer, llBufPoolPos, lpAddrInFile, llBufFileLen )
    endif
  endif
end


sub lReadFile( long lpHandle; var string spvBuffer$; long lpSize; var long lpvNumBytesRead )
  byte blIsAvailable, blIsOk, blHandleOk
  long llFiFlags, llFiFatEntryPos, llFiFilePointer, llBufPoolPos, &
    llBufFileStart, llBufFileLen, llReadFromBuf, llCurCluster, llCurAddr, &
    llCurSize, llAddrInFile, llCheckedSize, llPosition
  string slPoolBuffer$(FILE_DATA_BUFFER_SIZE)
  llReadFromBuf = 0
  slPoolBuffer$ = ""
  lpvNumBytesRead = -1
  call bCheckFileHandle( lpHandle, blHandleOk )
  if blHandleOk = TRUE then
    '' if lpSize is negative, return an error; if lpSize is zero, return zero
    if lpSize <= 0 then
      if lpSize = 0 then
        lpvNumBytesRead = 0
      endif
      return
    endif
    call lParseFileIndex( asFileIndexTable$( lpHandle ), llFiFlags, llFiFatEntryPos,&
      llFiFilePointer, llBufPoolPos, llBufFileStart, llBufFileLen )
    llPosition = llFiFilePointer
    '' check write-only flag
    '' combination of O_RDONLY and O_WRONLY is allowed
    if ( llFiFlags bitand O_WRONLY ) = O_WRONLY and &
      ( llFiFlags bitand O_RDONLY ) <> O_RDONLY then
        return
    endif
    call bFATGetDirEntry( llFiFatEntryPos, blIsAvailable )
    '' if end-of-file was reached, return zero
    if llFiFilePointer = lDirSize then
      lpvNumBytesRead = 0
      return
    endif
    llCheckedSize = lpSize
    if llCheckedSize > lDirSize - llPosition then
      llCheckedSize = lDirSize - llPosition
    endif
    '' read from the pool as many bytes as possible
    if llBufFileStart <= llPosition and llPosition < llBufFileStart + llBufFileLen then
      llReadFromBuf = llCheckedSize
      if llPosition + llCheckedSize > llBufFileStart + llBufFileLen then
        llReadFromBuf = llBufFileLen - ( llPosition - llBufFileStart )
      endif
      slPoolBuffer$ = stos$( slPoolBuffer$, 0, asFileDataPool$( llBufPoolPos ),&
        llPosition - llBufFileStart, llReadFromBuf )
    endif
    '' read the remained data from the medium directly
    if llReadFromBuf <> llCheckedSize then
      '' look up the cluster chain to find to which cluster the current address belongs
      llAddrInFile = llPosition + llReadFromBuf
      llCurCluster = wDirCluster
      call lFATFindSectorByAddr( llAddrInFile, llCurCluster, llCurAddr )
      llCurAddr = llCurAddr + mod( llPosition + llReadFromBuf, lBytesPerCluster )
      llCurSize = llCheckedSize - llReadFromBuf
      call sHdReadString( llCurAddr, llCurSize, spvBuffer$, blIsOk )
      if slPoolBuffer$ <> "" then
        spvBuffer$ = stos$( spvBuffer$, len(slPoolBuffer$), spvBuffer$, llCurSize )
        spvBuffer$ = stos$( spvBuffer$, 0, slPoolBuffer$, len(slPoolBuffer$) )
      endif
      llReadFromBuf = llCheckedSize
    else
      spvBuffer$ = slPoolBuffer$
    endif
    set_len$( spvBuffer$, llReadFromBuf )
    '' is the next position after the last read byte inside of file?
    '' yes - read the bytes from the next position into the pool
    if llPosition + llReadFromBuf < lDirSize then
      llCurAddr = llPosition + llReadFromBuf
      call bReadFileToDataPool( lpHandle, llCurAddr, blIsAvailable )
    endif
    call lSetFilePointer( lpHandle, lpSize, FILE_CURRENT, llPosition )
    lpvNumBytesRead = llReadFromBuf
  endif
end
#endif
#endcm


sub vInitFileFindVars()
  sFfpSearchedFileName$ = ""
  lFfpLastFoundPos = 0
end


'' for the using in the subroutine "bFindNextFile"
'' the file name to search for and the address at which the file was
'' actually found are saved in global vars
sub bSearchForFfpFile( var string spvFfdStruct$; var byte bpvFound )
  byte blIsAltNameAvailable, blIsOk
  long llFileInfoAddr, llLongNameSize
  string slLongFileName$(MAX_LONGNAME_SIZE)

#ifdef FS_SECURE_INIT
  slLongFileName$ = FS_STR_VARS_INIT_VALUE	'$JK
#endif	'' FS_SECURE_INIT

  bpvFound = FALSE
  '' look for the filename in the FAT
  call lFATFindFileInfo( sFfpSearchedFileName$, lFfpLastFoundPos, llFileInfoAddr )
  '' does the filename exist?
  if llFileInfoAddr >= 0 then
    call bFATGetDirEntry( llFileInfoAddr, blIsOk )
    if blIsOk = TRUE then
      spvFfdStruct$ = ""
      spvFfdStruct$ = ntos$( spvFfdStruct$, FFD_ATTR_OFFS, FFD_ATTR_SIZE, bDirAttr )
      spvFfdStruct$ = ntos$( spvFfdStruct$, FFD_CREATE_TIME_MS_OFFS, FFD_CREATE_TIME_MS_SIZE, bDirCreate10MS )
      spvFfdStruct$ = ntos$( spvFfdStruct$, FFD_CREATE_TIME_OFFS, FFD_CREATE_TIME_SIZE, wDirCreateTime )
      spvFfdStruct$ = ntos$( spvFfdStruct$, FFD_CREATE_DATE_OFFS, FFD_CREATE_DATE_SIZE, wDirCreateDate )
      spvFfdStruct$ = ntos$( spvFfdStruct$, FFD_ACCESS_DATE_OFFS, FFD_ACCESS_DATE_SIZE, wDirAccessDate )
      spvFfdStruct$ = ntos$( spvFfdStruct$, FFD_SIZE_OFFS, FFD_SIZE_SIZE, lDirSize )
      spvFfdStruct$ = stos$( spvFfdStruct$, FFD_NAME_OFFS, sDirName$, FFD_NAME_SIZE )
      spvFfdStruct$ = stos$( spvFfdStruct$, FFD_EXT_OFFS, sDirName$, FFD_EXT_OFFS, FFD_EXT_SIZE )
      if max_len( spvFfdStruct$ ) > FFD_STRUCT_SHORT_SIZE then
        call bFATGetLongNameLFN( llFileInfoAddr, slLongFileName$, MAX_LONGNAME_SIZE, blIsAltNameAvailable )
        if blIsAltNameAvailable = TRUE then
          llLongNameSize = len( slLongFileName$ )
          if llLongNameSize > max_len( spvFfdStruct$ ) - FFD_LONG_NAME_OFFS then
            llLongNameSize = max_len( spvFfdStruct$ ) - FFD_LONG_NAME_OFFS
          endif
          if llLongNameSize > FFD_LONG_NAME_SIZE then
            llLongNameSize = FFD_LONG_NAME_SIZE
          endif
      	  spvFfdStruct$ = stos$( spvFfdStruct$, FFD_LONG_NAME_OFFS, slLongFileName$, llLongNameSize )
        else
          if max_len( spvFfdStruct$ ) >= FFD_STRUCT_ABRIDGED_SIZE then
            slLongFileName$ = trim$( left$( sDirName$, FFD_NAME_SIZE ), " ", -1 )
            if len( trim$( mid$( sDirName$, FFD_NAME_SIZE, FFD_EXT_SIZE ), " ", -1 ) ) > 0 then
              slLongFileName$ = slLongFileName$ + DIR_EXT_TERM + &
                trim$( mid$( sDirName$, FFD_NAME_SIZE, FFD_EXT_SIZE ), " ", -1 )
            endif
            llLongNameSize = len( slLongFileName$ )
            spvFfdStruct$ = stos$( spvFfdStruct$, FFD_LONG_NAME_OFFS, slLongFileName$, llLongNameSize )
          endif
        endif
      endif
      lFfpLastFoundPos = llFileInfoAddr
      bpvFound = TRUE
    endif
  endif
end

#endif

